include <system.dh>
include <LookUp2DInc.dh>

const FE__CSYMB     502
const FE_STREET     60160
const AT_ROADID     1711

const delta         0.5

local fes 0
local ids 0

local error(msg)
{
  ArFreeAll();
  puts(msg);
  exit(0);
}

local getMeters(slat,slon,lat,lon)
{
  lat-=slat;
  lon-=slon;
  lon*=cos(deg2rad(slat));
  return sqrt(lat*lat+lon*lon)*111120; // 1852*60
}

local getMetersToNode(slat,slon,ch,nd)
{
  chGetNodeXY(ch,nd,lat,lon);
  return getMeters(slat,slon,lat,lon);
}

local moveNode(ch,nd,zlevel)
{
  prev=chSetGeoUnits(0);
  chGetNodeXY(ch,nd,x,y);
  chEdiNode(ch,nd,x,y+zlevel);
  chSetGeoUnits(prev);
}

local embedChart(chSrc,chDst)
{
  chPrealloc(chDst,chNodeNum(chSrc),chEdgeNum(chSrc),chLineNum(chSrc),chAreaNum(chSrc),chFeatureNum(chSrc));
  for( max=nd=0; (nd=chNextNode(chSrc,nd)); )
    if( max < nd )
      max=nd;
  cvrtNd=ArCreate(max);
  for( nd=1; nd<=max; nd+=1 )
    ArSetAt(cvrtNd,nd,0);
  for( max=eg=0; (eg=chNextEdge(chSrc,eg)); )
    if( max < eg )
      max=eg;
  cvrtEg=ArCreate(max);
  for( eg=1; eg<=max; eg+=1 )
    ArSetAt(cvrtEg,eg,0);
  for( nd=0; (nd=chNextNode(chSrc,nd)); )
    if( chGetNodeXY(chSrc,nd,lat,lon) )
      ArSetAt(cvrtNd,nd,chCrtNode(chDst,lat,lon/*+n360*/));
  for( eg=0; (eg=chNextEdge(chSrc,eg)); )
  {
    if( (sz=chGetEdgeBegEnd(chSrc,eg,beg,end)-2) >= 0 )
    {
      beg=ArGet(cvrtNd,beg);
      end=ArGet(cvrtNd,end);
      ArSetAt(cvrtEg,eg,egdst=chCrtEdge(chDst,beg,end));
      for( i=1; i<=sz; i+=1 )
        if( chGetEdgeXY(chSrc,eg,i,lat,lon) )
          chIncEdge(chDst,egdst,i,lat,lon/*+n360*/);
      if( (sz+=2) == chGetEdgeXYZ(chSrc,eg,0,lat,lon,hei) && chToggle2D3D(chDst,egdst) )
      {
        chEdiEdge(chDst,egdst,0,lat,lon/*+n360*/,hei);
        for( i=1; i<sz; i+=1 )
          if( chGetEdgeXYZ(chSrc,eg,i,lat,lon,hei) )
            chEdiEdge(chDst,egdst,i,lat,lon/*+n360*/,hei);
      }    
    }
  }
  for( max=feSrc=0; (feSrc=chNextFeature(chSrc,feSrc)); )
    if( max < feSrc )
      max=feSrc;
  cvrtFe=ArCreate(max);
  for( feSrc=0; (feSrc=chNextFeature(chSrc,feSrc)); )
  {
    fecode=chGetFeatureCode(chSrc,feSrc);
    feDst=chCrtFeature(chDst,fecode,chGetFeatureDictionary(chSrc,feSrc));
    ArSetAt(cvrtFe,feSrc,feDst);
    for( i=chGetAttrsNum(chSrc,feSrc); i; i-=1 ) 
    {
      atcode=chGetAttrCode(chSrc,feSrc,i);
      if( chGetAttrText(chSrc,feSrc,atcode,txt) )
      {
        if( dsAttrType(64003,atcode) == 2 && txt == "0" )
          continue;
        chSetAttrText(chDst,feDst,atcode,txt);
      }
    }
    if( (type=chGetMetricType(chSrc,feSrc)) == M_POINT )
    {
      if( chGetFeatureNode(chSrc,feSrc,nd) )
        chSetFeatureNode(chDst,feDst,ArGet(cvrtNd,nd));
    }
    else if( type == M_LINE )
    {
      if( (sz=chGetFeatureLine(chSrc,feSrc,1,eg,flag)) )
        chIncFeatureLine(chDst,feDst,1,ArGet(cvrtEg,eg),flag);
      for( i=2; i<=sz; i+=1 )
      {
        chGetFeatureLine(chSrc,feSrc,i,eg,flag);
        chIncFeatureLine(chDst,feDst,i,ArGet(cvrtEg,eg),flag);
      }
    }
    else if( type == M_AREA )
    {
      cntnum=chGetFeatureArea(chSrc,feSrc,0,0,null,null);
      for( i=1; i<=cntnum; i+=1 )
      {
        sz=chGetFeatureArea(chSrc,feSrc,i,1,eg,flag);
        for( j=1; j<=sz; j+=1 )
        {
          chGetFeatureArea(chSrc,feSrc,i,j,eg,flag);
          chIncFeatureArea(chDst,feDst,i,j,ArGet(cvrtEg,eg),flag);
        }
      }
    }
  }    
  for( feSrc=0; (feSrc=chNextFeature(chSrc,feSrc)); )
  {
    feDst=ArGet(cvrtFe,feSrc);
    sz=chGetSlaves(chSrc,feSrc,0); 
    for( i=1; i<=sz; i+=1 )
      chLinkM2S(chDst,feDst,ArGet(cvrtFe,chGetSlaves(chSrc,feSrc,i)));
    sz=chGetPeers(chSrc,feSrc,0); 
    for( i=1; i<=sz; i+=1 )
      if( (fe=chGetPeers(chSrc,feSrc,i)) > feSrc )
        chLinkP2P(chDst,feDst,ArGet(cvrtFe,fe));
  }
  ArFree(cvrtFe);
  ArFree(cvrtEg);
  ArFree(cvrtNd);
  chPrealloc(chDst);
  chFitBorder(chDst);
}  

local get1st(ids,roadid)
{
  if( !(fnd=ArBSearch(ids,roadid)) )
    return 0;
  while( (fnd-=1) )
  {
    if( ArGet(ids,fnd) != roadid )
      return fnd+1;
  }  
  return 1;
}

local getNodeByID(roadid,slat,slon,ids,fes,ch,feSrc)
{
  if( !(fnd=get1st(ids,roadid)) )
  {
    puts("* cannot find Road ID: "+roadid);
    return 0;
  }
  onvertex=0;
  conedge=0;  
  minnode=0;  
  mindist=delta;
  for( ; fnd<=ArSize(ids); fnd+=1 )
  {  
    if( ArGet(ids,fnd) != roadid )
      break;
    fe=ArGet(fes,fnd);  
    if( !chGetAttrValue(ch,fe,AT_ROADID,id) || id != roadid ) 
      error("*** failed Road ID: "+roadid+" != "+id);
    for( egn=chGetFeatureLine(ch,fe,1,eg,flag); egn; egn-=1 )
    {
      chGetFeatureLine(ch,fe,egn,eg,flag);
      len=chGetEdgeBegEnd(ch,eg,beg,end)-1;
      if( (distbeg=getMetersToNode(slat,slon,ch,beg)) < mindist )
      {
        mindist=distbeg;
        minnode=beg;
        if( beg == end )
          conedge=eg;
        else
          conedge=0;  
      } 
      if( (distend=getMetersToNode(slat,slon,ch,end)) < mindist )
      {
        mindist=distend;
        minnode=end;
        if( beg == end )
          conedge=eg;
        else
          conedge=0;  
      }
      if( !onvertex && !minnode )
      {
        while( (len-=1) > 0 )
        {
          chGetEdgeXY(ch,eg,len,lat,lon);
          if( getMeters(slat,slon,lat,lon) < delta )
          {
            onvertex=1;
            break;
          }  
        }
      } 
    }   
  }
  if( conedge )
  {
    prev=chSetGeoUnits(0);
    chGetNodeXY(ch,minnode,x,y);
    chSetEdgeEnd(ch,conedge,end=chCrtNode(ch,x,y));
    chSetGeoUnits(prev);
  } 
  if( !minnode && !onvertex )
    puts("* cannot find proper node of road id#"+roadid+" (zlevel symbol #"+feSrc+")");
  return minnode; 
}

global main()
{
  if( !(chSrc=GetEditDataset()) )
    return;
  chSetGeoUnits(1);  
  if( !chGetProperty(chSrc,CH_NAME,chname) ||
      !chGetProperty(chSrc,GEO_NORTH,no) ||
      !chGetProperty(chSrc,GEO_WEST,we) ||
      !chGetProperty(chSrc,GEO_SOUTH,so) ||
      !chGetProperty(chSrc,GEO_EAST,ea) ||
      !chGetProperty(chSrc,GEO_SCALE,scale) )
    return;
  if( !StrICmp(StrRight(chname,4),".dcf") )
    chname=StrLeft(chname,StrLen(chname)-4);
  if( !(ch=chCrtMercatorDataset(chname+"N.dcf",64003,scale,no,we,int((no+so)*0.5),"W84",1000)) )
    return;
  embedChart(chSrc,ch);  
  CloseDataset(chSrc);
  Load2Editor(ch);
  chUpdatePresentation(ch); //*/ ch=chSrc;
  
  for( chSrc=0; (chSrc=NextDataset(chSrc)) && chSrc == ch; );  
  if( !chSrc )
    error("*** cannot find zlevel dataset");
    
  fes=ArCreate(0);
  ids=ArCreate(0);
  for( fe=0; fe=chNextFeature(ch,fe); )
  {
    if( chGetFeatureCode(ch,fe) != FE_STREET )
      continue;
    if( !chGetAttrValue(ch,fe,AT_ROADID,id) ) 
      continue;
    ArAdd(fes,fe);    
    ArAdd(ids,id);    
    //if( 1 != chGetFeatureLine(ch,fe,0,eg,flag) )
    //  puts("multiline street #"+fe);
  }
//error("");
  if( !ArSize(fes) )
    error("*** wrong target dataset");
  ArSort(ids,fes);
  ArSort(ids,ids);
  done=ArCreate(0);
  tot=0;
  StartIndicator("Processing ...",chFeatureNum(chSrc));
  for( feSrc=0; feSrc=chNextFeature(chSrc,feSrc); )
  {
    StepIndicator();
    if( chGetFeatureCode(chSrc,feSrc) != FE__CSYMB || !chGetFeatureNode(chSrc,feSrc,nd) || !chGetNodeXY(chSrc,nd,slat,slon) )
      error("*** wrong zlevel dataset (none CSYMB): #"+feSrc);
    if( !chGetAttrText(chSrc,feSrc,SORIND,zlevel) || (zlevel=int(zlevel)) == 0 )
      continue; 
    if( !chGetAttrText(chSrc,feSrc,INFORM,roadid) || (roadid=int(roadid)) == 0 )
      error("*** wrong zlevel dataset (no Road ID): #"+feSrc);
    if( !(nd=getNodeByID(roadid,slat,slon,ids,fes,ch,feSrc)) )
      continue;
    moveNode(ch,nd,zlevel);
    ArAdd(done,nd);
    tot+=1;  
/* =====================================================================================================
    if( !(fnd=ArBSearch(ids,roadid)) )
    {
      puts("* cannot find Road ID: "+roadid);
      continue;
    }
    fe=ArGet(fes,fnd);  
    if( !chGetAttrValue(ch,fe,AT_ROADID,id) || id != roadid ) 
      error("*** failed Road ID: "+roadid+" != "+id);
    for( egn=chGetFeatureLine(ch,fe,1,eg,flag); egn; egn-=1 )
    {
      chGetFeatureLine(ch,fe,egn,eg,flag);
      len=chGetEdgeBegEnd(ch,eg,beg,end)-1;
      if( !(tst=chGetEdgesOfNode(ch,beg)) )
        error("*** orphan node #"+beg);
      else if( 1 != ArSize(tst) )
      {
        if( beg != end || ArSize(tst) != 2 || ArGet(tst,1) != ArGet(tst,2) )
          error("*** linked street #"+fe+" on node #"+beg);
        else
          puts("* street #"+fe+" closed on node #"+beg);  
      }  
      ArFree(tst);  
      if( !(tst=chGetEdgesOfNode(ch,end)) )
        error("*** orphan node #"+end);
      else if( 1 != ArSize(tst) )
      {
        if( beg != end || ArSize(tst) != 2 || ArGet(tst,1) != ArGet(tst,2) )
          error("*** linked street #"+fe+" on node #"+end);
      }  
      ArFree(tst);  
      if( (distbeg=getMetersToNode(slat,slon,ch,beg)) < delta )
      {
        if( beg == end )
        {
          prev=chSetGeoUnits(0);
          chGetNodeXY(ch,beg,x,y);
          chSetEdgeEnd(ch,eg,end=chCrtNode(ch,x,y));
          chSetGeoUnits(prev);
        }  
        else if( (distend=getMetersToNode(slat,slon,ch,end)) < delta )
        {
          if( distend < distbeg )
            beg=end;
          puts("* both nodes of street #"+fe);
        }
        moveNode(ch,beg,zlevel);
        ArAdd(done,beg);
        tot+=1;  
      }
      else if( (distend=getMetersToNode(slat,slon,ch,end)) < delta )
      {
        moveNode(ch,end,zlevel);
        ArAdd(done,end);
        tot+=1;  
      }
      else
      {
puts("street #"+fe+" distbeg:"+printf("%.3f",distbeg)+" distend:"+printf("%.3f",distend));
                             puts("  src node#"+ nd+printf(" [%.6f",slat)+printf(":%.6f]",slon));
chGetNodeXY(ch,beg,lat,lon); puts("  beg node#"+beg+printf(" [%.6f",lat)+printf(":%.6f]",lon));
chGetNodeXY(ch,end,lat,lon); puts("  end node#"+end+printf(" [%.6f",lat)+printf(":%.6f]",lon));
        while( (len-=1) > 0 )
        {
          chGetEdgeXY(ch,eg,len,lat,lon);
          if( getMeters(slat,slon,lat,lon) < delta )
            break;
        }
        if( !len )
          error("*** cannot find proper node of street #"+fe);
      }
    }
========================================================================================== */    
  }  
  ArSort(done,done);
  
  StartIndicator("Testing ...",ArSize(done)*2);
  chSetGeoUnits(0);
  for( i=ArSize(done); i>1; i-=1 )
  {
    StepIndicator();
    if( ArGet(done,i) == ArGet(done,i-1) )
      puts("** node #"+ArGet(done,i)+" moved several times");
  }     
  while( ArSize(done) )
  {
    StepIndicator();
    chGetNodeXY(ch,nd=ArGet(done,1),x0,y0);
    del=0;
    for( i=ArSize(done); i>1; i-=1 )
    {
      chGetNodeXY(ch,ArGet(done,i),x1,y1);
      if( x1 == x0 && y1 == y0 )
      {
        StepIndicator();
        ArRemoveAt(done,i);
        del+=1;
      }
    }
    if( !del )
      puts("** node #"+nd+" has no partner");
    ArRemoveAt(done,1);
  }
  ArFreeAll();
  FinishIndicator();
  puts("Total: "+tot);
  Redraw();
}
